// forint.hh, basic class of formula interpreter
/****************************************************************************

    C-formula interpreter library forint
    Copyright (C) 2000 Alexey Filin

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
    02111-1307, USA.
              
 ****************************************************************************/


#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>    // dynamic linking
#include <sys/types.h>// wait
#include <sys/wait.h> // wait
#include <iostream>
#include <strstream>
#include <string>
#include <vector>
#include <map>

#ifndef _FORINT_HH_
#define _FORINT_HH_

#include <forint/forint_types.hh>
#include <forint/stack_vec_poin_off.hh>
#include <forint/mssg.hh>
#include <forint/config.h>

namespace FInt {

struct fi_version_t {
  const char* name;
  int major_version;
  int minor_version;
  int release;
};

namespace FInt_internal {
extern const uint num_coor;
extern const char* cmd [];
}

class forint {

int locale;

// a formula
char* math_expression;

// flag of accessability of built-in variables
bool flag_of_built_in_var_access;

// flag of accessability of rl and self
bool flag_of_extra_built_in_var_access;

angle_unit_t angle_unit;

// flag of accessability of control (`while') statement
bool flag_of_ctrl_statement_access;

// flag of clearing built-in variables for beginning of the parse() call
bool clearing_flag;

// used by parse()
bool is_while;

char* get_C_string(char*, uint);
int get_size_0_string(char*, uint&);
void skip(char*&);
void rev_skip(char*&);
int check_name(const char*, const char*, bool, bool&);
int check_type(const char*&, b_type_t&, uint&, const char*);
int pair(char*&, uint, char, char);
int rev_pair(char*, char*&, char, char);
bool iftoken (char*, char*, char*&);
bool ntoken (char*, uint, char*&);
bool is_op_delim(char);
bool optoken (char*, uint, char*, char*&);
action_t token (char*, uint, action_t, char*, char*, char*&, uint&);
action_t token (char*, uint, action_t, action_t, char*, char*, char*&);
char* get_non_zero_min(char*, char*);
action_t find_token(uint, char*, uint, char*&, uint&);
action_t find_if_token(char*, char*, char*&, char*&,
		       char*&, char*&);
bool is_lvalue(char*, char*);
bool get_assignment(char*, char*&, uint&, action_t&);
bool find_assignment(char*, uint, char*&, uint&, char*&, uint&, action_t&);
int decompose(char*, char*, int, uint&);

int r_calculate(int&, double&, void*&);
int s_calculate(int, double&, int);
int dl_calculate(double&, char*);
double forint_dlf (void);

inline void
make_offsets(void)
{
  if (offsets_p->num) number_p += offsets_p->num;
  if (offsets_p->var) var_num_p += offsets_p->var;
  if (offsets_p->poin) poin_num_p += offsets_p->poin;
  if (offsets_p->poin_level) poin_num_level_p += offsets_p->poin_level;
  if (offsets_p->func0) usr_function_0_num_p += offsets_p->func0;
  if (offsets_p->func1) usr_function_1_num_p += offsets_p->func1;
  if (offsets_p->func2) usr_function_2d_num_p += offsets_p->func2d;
  if (offsets_p->func2p) usr_function_2p_num_p += offsets_p->func2p;
  if (offsets_p->func2) usr_function_2_num_p += offsets_p->func2;
  if (offsets_p->offsets) offsets_p += offsets_p->offsets;
}

void save_offsets(offsets_t& r);
void subtr_offsets(offsets_t& r);
void inverse_subtr_offsets(offsets_t& r);

double (*dl_function)(void); // dynamically linked function
void* dlfunc_lib;
char* dlfunc_cc_name;        // name of a  cc-file containing dl function
char* dlfunc_so_name;        // name of an so-file containing dl function

bool is_overloaded(const char*);
bool is_overloaded_func(const char*);
int mk_dl_func(char*,char**,char**,vector<char*>*);

mssg& c;

// possible coordinates -- x, y, z, phi, psi, r(polar), rad(spherical),
//                         t (2D parametric), u, v (3D parametric),
// recursion level      -- rl,
// possible constants   -- e, log_2 e, log_10 e, log_e 2, log_e 10,
//   pi, pi/2, pi/4, 1/pi, 2/pi, 2/sqrt(pi), sqrt(2), 1/sqrt(2)...
double argument [12]; // first is dummy

// here and later *_p pointers need only for some acceleration...
vec_t<action_t> r_action;
action_t* r_action_p;
vec_t<action_t> s_action;
action_t* s_action_p;

vec_t<double> number;
double* number_p;

vec_t<double*> variable;
double** variable_p;
vec_t<uint> var_num;
uint* var_num_p;
vec_t<char*> var_name;
vec_t<bool> var_const_flag;

vec_t<void*> pointer;
void** pointer_p;         // don't change it after an initiation !!!
vec_t<poin_t> poin_type;  // to check for a parsing
vec_t<uint*> poin_offset;
vec_t<uint*> poin_dim_size;
vec_t<uint> poin_num;
uint* poin_num_p;
vec_t<uint> poin_num_level;
uint* poin_num_level_p;  // don't change it after an initiation !!!
vec_t<char*> poin_name;

vec_t<USR_FUNC_0_PTR> usr_function_0;
USR_FUNC_0_PTR* usr_function_0_p;    // don't change it after an initiation !!!
vec_t<uint> usr_function_0_num;
uint* usr_function_0_num_p;
vec_t<char*> usr_function_0_name;

vec_t<USR_FUNC_1_PTR> usr_function_1;
USR_FUNC_1_PTR* usr_function_1_p;    // don't change it after an initiation !!!
vec_t<uint> usr_function_1_num;
uint* usr_function_1_num_p;
vec_t<char*> usr_function_1_name;

vec_t<USR_FUNC_2D_PTR> usr_function_2d;
USR_FUNC_2D_PTR* usr_function_2d_p;  // don't change it after an initiation !!!
vec_t<uint> usr_function_2d_num;
uint* usr_function_2d_num_p;
vec_t<char*> usr_function_2d_name;

vec_t<USR_FUNC_2P_PTR> usr_function_2p;
USR_FUNC_2P_PTR* usr_function_2p_p;    // don't change it after an initiation !!!
vec_t<poin_t> usr_function_2p_poin_type; // to check for a parsing
vec_t<uint> usr_function_2p_num;
uint* usr_function_2p_num_p;
vec_t<char*> usr_function_2p_name;

vec_t<USR_FUNC_2_PTR> usr_function_2;
USR_FUNC_2_PTR* usr_function_2_p;    // don't change it after an initiation !!!
vec_t<poin_t> usr_function_2_poin_type; // to check for a parsing
vec_t<uint> usr_function_2_num;
uint* usr_function_2_num_p;
vec_t<char*> usr_function_2_name;

char** r_prog;
char** s_prog;

// offsets keeper
stack_t<offsets_t> offsets;
offsets_t* offsets_p;

// 'long long' casting of rvalue in dlf
vector<uint> b_cast_pos;
vector<uint> e_cast_pos;
char* b_decomposed_string;
void push_cast_pos(char*,char*);

// 'double' casting of rvalue of `/' operator in dlf
// (both operands of `/' operator may be of integer type and result in DLF mode
// will be casted to integer also in the case)
vector<uint> d_b_cast_pos;
vector<uint> d_e_cast_pos;
void d_push_cast_pos(char*,char*);

//  'long long' lvalue
// I use fundamental property of map -- alignment of its elements
map<uint, string>   new_name_pos;
void push_long_name_pos(char*,string);
void insert_long_operands(string&);

bool break_flag;

void out_of_bound(char* name, uint top, int index);

void clean(void);

public:

int error_code;
// counter of dl functions
static uint dl_func_cntr;

// angle unit
void
setAngleUnit(angle_unit_t aunit=FI_RADIAN)
{ error_code = 0; angle_unit = aunit; }

angle_unit_t
getAngleUnit(void) const
{ return angle_unit; }

// name of file dynamically linked function is kept in
const char*
getCCName(void) const { return dlfunc_cc_name; }
const char*
getSOName(void) const { return dlfunc_so_name; }

forint (mssg& mes) : c(mes)
{
  break_flag = false;
  error_code = 0; 
  math_expression = 0;
  dl_function    = 0;
  dlfunc_lib     = 0;
  dlfunc_cc_name = 0;   // name of a  cc-file containing dl function
  dlfunc_so_name = 0;   // name of an so-file containing dl function
  dl_func_cntr   = 0;
  r_prog         = 0;
  s_prog         = 0;
  angle_unit     = FI_RADIAN;
// clear built-in variables
  for (uint i=0; i<FInt_internal::num_coor-1; i++)
    argument[i] = 0;
  argument[FI_RL] = 1;
  locale = c.getLocale();
}

virtual
~forint (void) { del(); }


// delete all captioned dynamic memory and close DLF-lib
void del(void);

double& x   (void) { return argument[FI_X]; }
double& y   (void) { return argument[FI_Y]; }
double& z   (void) { return argument[FI_Z]; }
double& phi (void) { return argument[FI_PHI]; }
double& psi (void) { return argument[FI_PSI]; }
double& r   (void) { return argument[FI_R]; }
double& rad (void) { return argument[FI_RAD]; }
double& t   (void) { return argument[FI_T]; }
double& u   (void) { return argument[FI_U]; }
double& v   (void) { return argument[FI_V]; }
double& rl  (void) { return argument[FI_RL]; }

int putConst(const double& var, const char* name, bool force=false);
int putVar  (double& var, const char* name, bool force=false);
int putPoin (void*   ptr, const char* name, const char* type,
	     vector<int>* dim_size=0, bool force=false);
int putFunc0 (USR_FUNC_0_PTR ptr, const char* name, bool force=false);
int putFunc1 (USR_FUNC_1_PTR ptr, const char* name, bool force=false);
int putFunc2D(USR_FUNC_2D_PTR ptr,const char* name, bool force=false);
int putFunc2P(USR_FUNC_2P_PTR ptr,const char* name,
	      const char* type1, const char* type2, bool force=false);
int putFunc2 (USR_FUNC_2_PTR ptr, const char* name,
	      const char* type, bool force=false);

// make parsing of expr in mode:
//   NEW_FORM_MODE -- make overfull clearing
//   ADD_FORM_MODE -- add expression to previous one (if one)
int parse(char* expr,
	  parse_mode_t mode = NEW_FORM_MODE,
	  bool bvar_access  = true,
	  bool ebvar_access = true,
	  bool ctrl_access  = true,
	  bool clear = false);

// make DL function, return zero for a successful completing
int makeDLF(vector<char*>* dlf_opt=0);

// make calculation by formula in a mode:
//   RVM_MODE - by a recurcive calls virtual machine,
//   SVM_MODE - by a stack virtual machine,
//   DLF_MODE - by a dynamic linking of a compiled formula, in the mode
//              external names must coinside with their aliases !!! ...
double calculate(calc_mode_t);

// virtual machines programm
char** getRProg(int& length);
char** getSProg(int& length);
int getRProgSize(void) { return r_action.size(); }
int getSProgSize(void) { return s_action.size(); }

// clear built-in variables
void
clear(void)
{
  for (uint i=0; i<FInt_internal::num_coor-1; i++)
    argument[i] = 0;
  argument[FI_RL] = 1;
}

void
getVersion(fi_version_t& ver)
{
  ver.name = FORINT_PACKAGE;
  ver.major_version = (FORINT_SEQ_VERSION & 0xFF0000) >> 16;
  ver.minor_version = (FORINT_SEQ_VERSION & 0xFF00) >> 8;
  ver.release = FORINT_SEQ_VERSION & 0xFF;
}

// get pointer to break_flag
bool*
getPBrk(void) { return &break_flag; }

  int copy(const forint& t);
  forint(const forint& t);
// ATTENTION: mssg& doesn't copied !!!
  forint& operator = (const forint& t);

};

}

#endif // _FORINT_HH_
